// -*- Mode:C++ -*-
//Header:
//File: aheap.h
//Title: array heap template
//Author: NODA Itsuki
//Date: 1998/3/25
//EndHeader:

//ModifyHistory:
// 1998/3/25: Create this file
//EndModifyHistory:

//Memo:
//EndMemo:

#ifndef _aheap_h_
#define _aheap_h_
////////////////////////////////////////////////////////////////////////

#include "base.h"

//======================================================================
// AHeap<Content>
//
#ifndef DEFAULT_ARRAY_SIZE
#define DEFAULT_ARRAY_SIZE	256
#endif
#ifndef DEFAULT_EXPAND_MAG
#define DEFAULT_EXPAND_MAG	2
#endif


template <class _Contents_>
class AHeap {
   public:
      //--------------------------------------------------
      // LocalTypeDef
      typedef AHeap<_Contents_> aheap_t ; 
      typedef _Contents_ cont_t ;

      cont_t** table ;
      UInt _top ;
      UInt size ;
      UInt mag ;

      inline AHeap<_Contents_>(UInt sz = DEFAULT_ARRAY_SIZE,
			       UInt mg = DEFAULT_EXPAND_MAG) {
				  size = 0 ;
				  mag = mg ;
				  _top = 0 ; 
				  expand(sz) ;
      } ;
     
      inline bool expand() { return expand(size * mag) ; } ;

      inline bool expand(UInt newsize) {
#ifdef DEBUG0
	 cerr << "Status: expanding ArrayHeap:" << size << "->" << newsize
	      << endl ;
#endif	  
	 cont_t** oldtable = table ;
	 table = new cont_t*[newsize] ;
	 for(UInt i = 0 ; i < size ; i++) {
	    table[i] = oldtable[i] ;
	 }
	 UInt sizedif = newsize - size ;
	 cont_t* newbuf = new cont_t[sizedif] ;
	 for(UInt i = 0 ; i < sizedif ; i++) {
	    table[i+size] = &(newbuf[i]) ;
	 }
	 size = newsize ;
	 return true ;
      } ;

      inline cont_t* get() {
	 if(_top >= size) expand() ;
	 cont_t* r = table[_top] ;
	 r->init() ;
	 _top++ ;
	 return r ;
      } ;

      inline UInt toppos() { return _top ; } ;
     
      inline bool rewind(UInt pos) { _top = pos ; return true ; } ;

      inline friend ostream& operator<< (ostream& o, const aheap_t& hp) {
	 o << "#ArrayHeap[size=" << hp.size << ",top=" << hp._top 
	   << ",mag=" << hp.mag << "]" << endl ;
	 for(UInt i = hp._top ; i > 0 ; i--) {
	    o << "\t" << *(hp.table[i-1]) << endl ;
	 } 
	 o << "\t___bottom___" << endl ;
      } ;
};

////////////////////////////////////////////////////////////////////////
#endif
