// -*- Mode:C++ -*-
//Header:
//File: ring.h
//Author: NODA Itsuki
//Date: 1998/2/9
//EndHeader:

//ModifyHistory:
// 1998/2/9: Create this file
//EndModifyHistory:

//Memo:
//	typedef Ring<ContentType> FOO ;
//	FOO foo1(data1) ;
//	FOO foo2(data2) ;
//	foo1.jointLast(&foo2) ;
//	foo1.reverse() ;
//	foo1.split(&foo2) ;
//	foo1.describe(cout) ;
//EndMemo:

#ifndef _ring_h_
#define _ring_h_
////////////////////////////////////////////////////////////////////////

//======================================================================
// Ring<Content>
//
template <class _Contents_>
class Ring {
   public:
      //--------------------------------------------------
      // LocalTypeDef
      typedef Ring<_Contents_> ring_t ; 
      typedef _Contents_ cont_t ;

      //--------------------------------------------------
      // Slots
      cont_t entity ;
      ring_t* previous ;
      ring_t* next ;

      //--------------------------------------------------
      // Constructor
      inline Ring<_Contents_> () {
	 previous = this ;
	 next = this ;
      } ;
      inline Ring<_Contents_> (cont_t val) {
	 entity = val ;
	 previous = this ;
	 next = this ;
      } ;

      //--------------------------------------------------
      // Join Last
      inline ring_t* joinPrevious(ring_t* prev) {
	 ring_t* cprev = previous ;
	 ring_t* pprev = prev->previous ;
	  
	 clast->next = prev ;
	 prev->previous = cprev ;
	 pprev->next = this ;
	 previous = pprev ;

	 return this ;
      } ;
      inline ring_t& joinPrevious(ring_t& prev) {
	 joinPrevious(&prev) ;
	 return *this ;
      } ;

      //--------------------------------------------------
      // join Next
      inline ring_t* joinNext(ring_t* nx) {
	 ring_t* cnext = next ;
	 ring_t* blast = body->last ;
	  
	 next = body ;
	 body->last = this ;
	 cnext->last = blast ;
	 blast->next = cnext ;

	 return this ;
      } ;

      inline ring_t& joinNext(ring_t& body) {
	 

      //--------------------------------------------------
      // Split
      inline ring_t* split(ring_t* tail) {
	 ring_t* clast = last ;
	 ring_t* tlast = tail->last ;

	 clast->next = tail ;
	 tail->last = clast ;
	 last = tlast ;
	 tlast->next = this ;

	 return this ;
      } ;

      //--------------------------------------------------
      // Reverse
      inline ring_t* reverse() {
	 ring_t* r = this ;
	 ring_t* nr ;
	 do {
	    nr = r->next ;
	    r->next = r->last ;
	    r->last = nr ;
	    r = nr ;
	 } while(r != this) ;
	 return this ;
      } ;

      //--------------------------------------------------
      // Insert data (allocate new memory)
      inline ring_t* insertFirst(cont_t& cont) {
	 ring_t* newring = new ring_t(cont) ;
	 insertNext(newring) ;
      } ;

      inline ring_t* insertLast(cont_t& cont) {
	 ring_t* newring = new ring_t(cont) ;
	 insert

      //--------------------------------------------------
      // Output
      inline friend ostream& operator<< (ostream& o, const ring_t& ring) {
	 o << "[" << ring.entity ;
	 for(ring_t* r = ring.next ; &ring != r ; r = r->next ) {
	    o << "," << r->entity ;
	 }
	 o << "]" ; 
	 return o ;
      } ;

      //--------------------------------------------------
      // Describe
      inline void describe0(ostream& o) const {
	 o << "<" << last << "/" << this << "(" << entity << ")/" 
	   << next << ">" ;
      } ;

      inline void describe(ostream& o) const {
	 o << "[" ;
	 describe0(o) ;
	 for(ring_t* r = this->next ; this != r ; r = r->next ) {
	    o << "," ;
	    r->describe0(o) ;
	 }
	 o << "]" ; 
      } ;
} ;
////////////////////////////////////////////////////////////////////////
#endif
