// -*- Mode:C++ -*-
//Header:
//File: splayer.h
//Author: NODA Itsuki
//Date: 1998/6/23
//EndHeader:

//ModifyHistory:
// 1998/6/23: Create this file
//EndModifyHistory:

#ifndef _splayer_h_
#define _splayer_h_
////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include "base.h"
#include "gem.h"

extern "C" {
#include "libsclient.h" 
};

inline Pos xyP(Float x, Float y) { 
   Pos p ; p.x = x ; p.y = y ; return p ; 
};

//======================================================================
// DEFS
//
#define BUFFERSIZE	4096
#define BUFRINGSIZE	128

#define DASHPOWERFACTOR	0.02
#define PLAYERSIZE 0.8

#define FIELDLENGTH 105.0
#define FIELDHALFLENGTH (FIELDLENGTH/2.0)
#define FIELDWIDTH 68.0
#define FIELDHALFWIDTH (FIELDWIDTH/2.0)


//======================================================================
// util
//
inline Float normalizeAngle(Float ang) {
   while(ang < -180) ang += 360 ;
   while(ang > 180)  ang -= 360 ;
   return ang ;
} ;

#ifndef M_PI
#define M_PI        3.14159265358979323846	/* pi */
#endif
#define DEG2RAD	(M_PI/180.0)
#define RAD2DEG	(180.0/M_PI)

inline Float Sin(Float deg) { return sin(deg * DEG2RAD) ; };
inline Float Cos(Float deg) { return cos(deg * DEG2RAD) ; };
inline Float Atan2(Float y, Float x) { return RAD2DEG * atan2(y,x) ; } ;

inline Float sq(Float v) { return v * v ; } ;
inline Float dist(Pos a, Pos b) {
   return sqrt(sq(a.x - b.x) + sq(a.y - b.y)) ;
} ;

inline void dbp(Pos pos) { 
   cerr << "(" << pos.x << "," << pos.y << ")" ;
} ;

//======================================================================
// Buffer Ring Buffer
//
class HearInfoBuffer {
   public:
      HearInfo* bufRing ;
      Int ringsize ;
      Int top ;
      pthread_mutex_t mutex ;

      inline HearInfoBuffer(Int rsize = BUFRINGSIZE) {
	 ringsize = rsize ;
	 top = 0 ;
	 bufRing = new HearInfo[ringsize] ;
	 pthread_mutex_init(&mutex,NULL) ;
      } ;

      inline HearInfo* getBuf(Int n) {
	 return &(bufRing[n%ringsize]) ;
      } ;
      inline HearInfo* getBuf() {
	 return getBuf(top) ;
      } ;

      inline bool push(HearInfo* hinfo) {
	 pthread_mutex_lock(&mutex) ;
	 HearInfo* b = getBuf() ; 
	 memcpy(b,hinfo,sizeof(HearInfo)) ;
	 top++ ;
	 pthread_mutex_unlock(&mutex) ;
      } ; 
} ;
	  
//======================================================================
// ObjTable
//
class ObjInfo {
   public:
      SeenObj seen ;
      Pos apos ;
      Pos avec ;
      Float adir ;
      Float fdir ;
      Int etime ;
} ;
class ObjTable {
   public:
      ObjInfo teammate[11] ;
      ObjInfo opponent[11] ;
      ObjInfo ball ;
      ObjInfo goal ;
      ObjInfo owngoal ;

      inline bool init(Side side) {
	 for(Int i = 0 ; i < 11 ; i++) {
	    opponent[i].seen.type = OT_Player ;
	    opponent[i].seen.inViewConeP = (Bool)true ;
	    opponent[i].seen.id.player.team = WT_Opponent ;
	    opponent[i].seen.id.player.unum = i+1 ;
	    opponent[i].etime = -1 ;
	    teammate[i].seen.type = OT_Player ;
	    teammate[i].seen.inViewConeP = (Bool)true ;
	    teammate[i].seen.id.player.team = WT_Our ;
	    teammate[i].seen.id.player.unum = i+1 ;
	    teammate[i].etime = -1 ;
	 }
	 ball.seen.type = OT_Ball ;
	 ball.seen.inViewConeP = (Bool)true ;
	 ball.etime = -1 ;
	 goal.seen.type = OT_Goal ;
	 goal.seen.inViewConeP = (Bool)true ;
	 goal.seen.id.side = ((side == S_LEFT)?S_RIGHT:S_LEFT) ;
	 goal.etime = -1 ;
	 owngoal.seen.type = OT_Goal ;
	 owngoal.seen.inViewConeP = (Bool)true ;
	 owngoal.etime = -1 ;
	 owngoal.seen.id.side = side ;
      } ;
} ;

//======================================================================
// Player
//

//class PosInfo ;

class PassInfo {
   public:
      enum KickType { KT_none, KT_pass, KT_clear, KT_shoot } ;
      Pos pos ;
      KickType type ;

      inline PassInfo() { pos.x = 0 ; pos.y = 0 ; type = KT_none ; } ;
      inline PassInfo(const Pos& p, KickType t) {
	 pos.x = p.x ; pos.y = p.y ; type = t ;
      } ;
      inline PassInfo(Float x, Float y, KickType t) {
	 pos.x = x ; pos.y = y ; type = t ;
      } ;
      inline PassInfo(const PassInfo& pinfo) {
	 pos.x = pinfo.pos.x ; pos.y = pinfo.pos.y ; type = pinfo.type ; 
      } ;
} ;

class Player {
   public:
      Gem* sensorProcess ;
      Gem* motorProcess ;
      Gem* watchdogProcess ;
      Gem* listenerProcess ;
      Gem* planningProcess ;

      Socket sock ;
      char* serverhost ;
      Int serverport ;

      char* sendBuffer ;
      char* recvBuffer ;
      Int bufferSize ;

      InitInfo initInfo ;
      char* teamname ;
      Side side ;
      Int unum ;
      PlayMode playmode ;

//      PosInfo* posInfo ;
//      PosInfo** passTo ;
      PassInfo* passTo ;
      UInt passToN ;
      UInt passToMax ;

      struct timeval startTime ;

      SensorInfo sensorInfo ;
      PosState currentPos ;
      UInt ctime ;
      CondBarrier ctimeBarrier ;
      UInt vtime ;
      UInt etime ;
      CondBarrier vtimeBarrier ;
      UInt atime ;
      CondBarrier atimeBarrier ;
      HearInfoBuffer messageBuffer ;
      UInt listendTop ;
      UInt sbtime ;
      CondBarrier sbtimeBarrier ;
      SenseBodyInfo sbInfo ;

      ObjTable objTable ;

      Int wKickOffTime ;

      bool VerboseP ;
      Float DashPower ;
      Float PredFactor ;
      Float ShiftFactor ;

      //--------------------------------------------------
      // Initialize
      inline Player() {
	 bufferSize = BUFFERSIZE ;
	 sendBuffer = new char[bufferSize] ;
	 recvBuffer = new char[bufferSize] ;
	 listendTop = 0 ;
	 VerboseP = false ;
	 PredFactor = 0.0 ;
	 DashPower = 70 ;
	 ShiftFactor = 1.0 ;
      } ;

      bool init(char* host, Int port, char* team, bool goalie) ;
      bool openConnection(char* host, Int port) ;

      //--------------------------------------------------
      // sending Command
      bool sendInit(char* teamname, bool goalie) ;
      inline bool sendMove(Float x, Float y) {
	 return send_com_move(&sock,x,y) ;
      } ;
      inline bool sendTurn(Float power, UInt sleep = 0, bool backp = false) {
	 if(fabs(power) < 1.0) return false ;
	 if(backp) { power = normalizeAngle(power+180) ; }
	 bool r = send_com_turn(&sock,power) ;
	 currentPos.dir += power ;
	 if(sleep > 0) MSleep(sleep) ;
	 return r ;
      } ;
      inline bool sendTurnAbs(Float dir, UInt sleep = 0, bool backp = false) {
	 return sendTurn(normalizeAngle(dir - currentPos.dir), sleep, backp) ;
      } ;
      inline bool sendTurnAbs(Pos p, UInt sleep = 0, bool backp = false) {
	 Float d = Atan2(p.y - currentPos.pos.y, p.x - currentPos.pos.x) ;
	 return sendTurnAbs(d, sleep, backp) ;
      } ;
      inline bool sendDash(Float power, UInt sleep = 0, bool backp = false) {
	 if(backp) power *= -1 ;
	 bool r = send_com_dash(&sock,power) ;
	 currentPos.pos.x += DASHPOWERFACTOR * power * Cos(currentPos.dir) ;
	 currentPos.pos.y += DASHPOWERFACTOR * power * Sin(currentPos.dir) ;
	 if(sleep > 0) MSleep(sleep) ;
	 return r ;
      } ; 
      inline bool sendKick(Float power, Float dir, UInt sleep = 0) {
	 bool r = send_com_kick(&sock,power,dir) ;
	 if(sleep > 0) MSleep(sleep) ;
	 return r ;
      } ;
      inline bool sendKickAbs(Float power, Float adir, UInt sleep = 0) {
	 return sendKick(power,normalizeAngle(adir - currentPos.dir), sleep) ;
      } ;
      inline bool sendKickAbs(Float power, Pos p, UInt sleep = 0) {
	 Float d = Atan2(p.y - currentPos.pos.y, p.x - currentPos.pos.x) ;
	 return sendKickAbs(power, d, sleep) ;
      } ;
      inline bool sendSay(char* message, UInt sleep = 0) {
	 bool r = send_com_say(&sock,message) ;
	 if(sleep > 0) MSleep(sleep) ;
	 return r ;
      } ;
      inline bool sendSenseBody(UInt sleep = 0) {
	 bool r = send_com_sense_body(&sock) ;
	 if(sleep > 0) MSleep(sleep) ;
	 return r ;
      } ;
      inline bool sendCatch(Float dir, UInt sleep = 0) {
	 bool r = send_com_catch(&sock,dir) ;
	 if(sleep > 0) MSleep(sleep) ;
	 return r ;
      } ;



      //--------------------------------------------------
      // receiving info
      inline bool recvInfo(bool firstp = true ) {
	 if(firstp) {
	    return receive_message(&sock, recvBuffer, bufferSize) ;
	 } else {
	    return receive_info(&sock, recvBuffer, bufferSize) ;
	 }
      } ;
      inline bool scanInfo() {
	 return scan_info(recvBuffer, &sensorInfo, teamname) ;
      } ;

      //--------------------------------------------------
      // estimate current pos
      inline bool estimateCurrentPos() {
	 return 
	    estimate_current_pos2(&(sensorInfo.info.see),side,&currentPos) ;
      } ;

      //--------------------------------------------------
      // estimate current pos

      bool updateAInfo() ;
      bool updateVInfo() ;
      bool updateSBInfo() ;

      bool updateObjTable() ;
      bool updateObjInfo(ObjInfo* oinfo, SeenObj* sobj) ;

      inline bool ballKickable() {
	 return dist(objTable.ball.apos,currentPos.pos) < 
	    (0.8 + 0.085 + 0.8) ;
	 //	       (0.8 + 0.085 + 1.0) ;
      } ;

      //--------------------------------------------------
      // wait new info
      inline bool waitVInfo() {
	 vtimeBarrier.lockwait() ;
	 return true ;
      } ;
      inline bool waitAInfo() {
	 atimeBarrier.lockwait() ;
	 return true ;
      } ;
      inline bool waitSBInfo() {
	 sbtimeBarrier.lockwait() ;
	 return true ;
      } ;

      //--------------------------------------------------
      // get time stamp
      inline UInt rtime() {
	 struct timeval ct ;
	 gettimeofday(&ct,NULLPTR) ;
	 return (ct.tv_sec - startTime.tv_sec) * 1000000 
	    + (ct.tv_usec - startTime.tv_usec) ;
      } ;
} ;

//======================================================================
// misc
//

inline bool beforeKickOffP(Player* player) {
   if(player->wKickOffTime >= 0) return true ;
   return false ;
}

//======================================================================
// Sensor Process
//

extern void* sensorLoop(Player *player) ;
extern void sensorCycle(Player *player) ;

//======================================================================
// Basic Actions
//

extern void splayerInit() ;

extern bool splay_get_new_message(Gem* gem) ;
extern bool splay_new_player(Gem* gem) ;

extern bool splay_send_move(Gem* gem) ;
extern bool splay_send_turn(Gem* gem) ;
extern bool splay_send_turn_adir(Gem* gem) ;
extern bool splay_send_turn_apos(Gem* gem) ;
extern bool splay_send_dash(Gem* gem) ;
extern bool splay_send_kick(Gem* gem) ;
extern bool splay_send_kick_adir(Gem* gem) ;
extern bool splay_send_kick_apos(Gem* gem) ;
extern bool splay_send_say(Gem* gem) ;
extern bool splay_send_sense_body(Gem* gem) ;
extern bool splay_send_catch(Gem* gem) ;

extern bool splay_get_unum(Gem* gem) ; 
extern bool splay_get_side(Gem* gem) ; 

extern bool splay_get_cpos(Gem* gem) ; /* current position */
extern bool splay_get_bpos(Gem* gem) ; /* ball position */
extern bool splay_get_ppos(Gem* gem) ; /* players position */
extern bool splay_get_ctime(Gem* gem) ;
extern bool splay_get_vtime(Gem* gem) ;
extern bool splay_get_atime(Gem* gem) ;
extern bool splay_get_stime(Gem* gem) ;
extern bool splay_get_rtime(Gem* gem) ;
extern bool splay_get_betime(Gem* gem) ;
extern bool splay_wait_vinfo(Gem* gem) ;
extern bool splay_wait_ainfo(Gem* gem) ;
extern bool splay_wait_sinfo(Gem* gem) ;

extern bool splay_sensor_loop(Gem* gem) ;

extern bool splay_chase_ball_simple(Gem* gem) ;
extern bool splay_chase_ball_better(Gem* gem) ;
extern bool splay_new_pass_dst(Gem* gem) ;
extern bool splay_clear_pass_dst(Gem* gem) ;
extern bool splay_add_pass_dst(Gem* gem) ;
extern bool splay_list_pass_dst(Gem* gem) ;
extern bool splay_chase_pass(Gem* gem) ;
extern bool splay_turn_look(Gem* gem) ;
extern bool splay_goto_pos(Gem* gem) ;

extern bool splay_inside_rectangle(Gem* gem) ;
extern bool splay_shifted_pos(Gem* gem) ;
extern bool splay_kickable(Gem* gem) ;

////////////////////////////////////////////////////////////////////////
#endif

     
